------------------------------ MODULE axi_base ------------------------------
EXTENDS Sequences,Integers,Naturals
CONSTANT DATA
CONSTANT DATA_DONT_CARE
CONSTANT ADDR_DONT_CARE
CONSTANT ID_DONT_CARE
CONSTANT ADDR
CONSTANT ID
CONSTANT MAX_TRAN_LEN
VARIABLE tran_seq
VARIABLE addr_ch_state
VARIABLE data_ch_state 
VARIABLES awvalid
VARIABLES awready,awid,awaddr
VARIABLES wvalid,wready,wid,wdata





addr_ch_vars == <<awvalid,awready,awid,awaddr>>
data_ch_vars == <<wvalid,wready,wid,wdata>>
vars == << tran_seq,addr_ch_state,data_ch_state >> \o addr_ch_vars \o data_ch_vars


\* Singles only 
\* Outstanding = 1


ASSUME MAX_TRAN_LEN \in Nat \ {0}

\* Channels
ADDR_WR_CH == INSTANCE axi_channel WITH Data <- ADDR,ID <- ID,valid <- awvalid,ready <- awready,data <- awaddr,id <- awid
DATA_WR_CH == INSTANCE axi_channel WITH Data <- DATA,ID <- ID,valid <- wvalid,ready <- wready,data <- wdata,id <- wid

DONT_CARE_TRAN_IDX == MAX_TRAN_LEN + 1



tranType == [
 dir : {"Wr"} \* TODO : Extend to read channel
,adr : ADDR
,data : DATA
,data_state : {"Idle" ,"Busy" ,"Done"} 
,addr_state : {"Idle" ,"Busy" ,"Done"} 
,tran_id : 1..MAX_TRAN_LEN 
,id : ID
]

tranDefault == [
  dir |-> "Wr"
 ,addr |-> CHOOSE x \in ADDR : TRUE
 ,data |-> CHOOSE x \in DATA : TRUE
 ,data_state |-> "Idle"
 ,addr_state |-> "Idle"
 ,tran_id |-> 1
 ,id |-> CHOOSE x \in ID : TRUE
]

transInit == [x \in 1..MAX_TRAN_LEN  |-> tranDefault ]
transInit2 == [x \in 1..MAX_TRAN_LEN |-> [transInit[x] EXCEPT !.tran_id  = x] ]

    
chanStateType == [
  state : {"Idle","Busy","Done"}
  ,tran_idx : 1..MAX_TRAN_LEN \cup {DONT_CARE_TRAN_IDX}
]    

TypeOK ==
    /\ \E x \in 1..MAX_TRAN_LEN : tran_seq \in [1..x -> tranType] 
    /\ addr_ch_state \in chanStateType
    /\ data_ch_state \in chanStateType
    /\ \A x,y \in DOMAIN tran_seq : ((tran_seq[x].tran_id # tran_seq[y].tran_id) \/ (x = y))

haveWaitingTran == 
    /\ Len(tran_seq) > 0
    /\ \E x \in DOMAIN tran_seq : tran_seq[x].state = "Idle"


MinAddrIdle(s) ==
    LET 
        idle == {i \in DOMAIN s : s[i].addr_state = "Idle"}
        res == CHOOSE min : \A y \in idle : min <= y
    IN res
  
MinDataIdle(s) ==
    LET 
        idle == {i \in DOMAIN s : s[i].data_state = "Idle"}
        res == CHOOSE min : \A y \in idle : min <= y
    IN res

\* Lets transaction sequence is initialized to some value
\* Lets define : Transaction in progress (TranProg) is transaction that has either channel state
\* per transaction busy or done but not both of them
TranProg(t) == TRUE \* TODO : Define it 

debugAddrChOnly(trans,a_ch_state) ==
    /\  LET 
            tran_prog == {x \in DOMAIN trans : trans[x].addr_state = "Idle" }
            res_set == {
                <<
                    [trans EXCEPT ![y].addr_state = "Busy"]
                    ,trans[y].id
                    ,trans[y].addr
                    ,[a_ch_state EXCEPT !.state = "Busy",!.tran_idx = y]
                    >> : y \in tran_prog  
                }
        IN res_set


StartAddrChOnly(trans,next_trans,next_addr,next_addr_id,a_ch_state,next_a_ch_state) == 
    /\ \E idx \in DOMAIN trans : trans[idx].addr_state = "Idle" 
    /\  LET 
            tran_prog == {x \in DOMAIN trans : trans[x].addr_state = "Idle" }
            res_set == {
                <<
                    [trans EXCEPT ![y].addr_state = "Busy"]
                    ,trans[y].id
                    ,trans[y].addr
                    ,[a_ch_state EXCEPT !.state = "Busy",!.tran_idx = y]
                    >> : y \in tran_prog  
                }
        IN <<next_trans,next_addr_id,next_addr,next_a_ch_state>> \in res_set  
        
         

\* Find 
StartDataChOnly(trans,next_trans,next_data,next_data_id,d_ch_state,next_d_ch_state) == 
    /\ \E idx  \in DOMAIN trans : trans[idx].data_state = "Idle" 
    /\  LET 
            tran_prog == {x \in DOMAIN trans : trans[x].data_state = "Idle" }
            res_set == {
                <<
                [trans EXCEPT ![y].data_state = "Busy"]
                ,trans[y].id
                ,trans[y].data
                ,[d_ch_state EXCEPT !.state = "Busy",!.tran_idx = y]
                >> : y \in tran_prog  
                }
        IN <<next_trans,next_data_id,next_data,next_d_ch_state>> \in res_set  

StartBothCh(trans,next_trans,next_data,next_addr,next_addr_id,next_data_id,a_ch_state,next_a_ch_state,d_ch_state,next_d_ch_state) ==
    /\  \E idx \in DOMAIN trans : trans[idx].addr_state = "Idle" 
    /\  \E idx \in DOMAIN trans : trans[idx].data_state = "Idle" 
    /\  LET 
            idle_dch ==  {idx \in DOMAIN trans : trans[idx].data_state = "Idle" } 
            idle_ach ==  {idx \in DOMAIN trans : trans[idx].addr_state = "Idle" }
            res_set == {
                <<
                [trans EXCEPT ![x].data_state = "Busy"
                ,![y].addr_state = "Busy"]
                ,trans[x].data
                ,trans[x].id
                ,trans[y].addr
                ,trans[y].id
                ,[a_ch_state EXCEPT !.state = "Busy",!.tran_idx = y]
                ,[d_ch_state EXCEPT !.state = "Busy",!.tran_idx = x]
                >> : x \in idle_dch,y \in idle_ach  }
        IN <<next_trans,next_data,next_data_id,next_addr,next_addr_id,next_a_ch_state,next_d_ch_state>> \in res_set

\* Update transactions when any channel is done
ChDone(trans,next_trans,a_ch_state,next_a_ch_state,d_ch_state,next_d_ch_state,ddone,adone) ==
    <<next_trans,next_a_ch_state,next_d_ch_state>> = LET 
      step1 == IF adone = TRUE THEN [trans EXCEPT ![a_ch_state.tran_idx].addr_state = "Done"] ELSE trans
      step2 == IF ddone = TRUE THEN [step1 EXCEPT ![d_ch_state.tran_idx].data_state = "Done"] ELSE step1
      ach_step == IF adone = TRUE THEN [a_ch_state EXCEPT !.state = "Done"] ELSE a_ch_state 
      dch_step == IF ddone = TRUE THEN [d_ch_state EXCEPT !.state = "Done"] ELSE d_ch_state
      res == <<step2,ach_step,dch_step>>
    IN res
    

Start ==
    \/ /\ awvalid = 0
       /\ StartAddrChOnly(tran_seq,tran_seq',awaddr',awid',addr_ch_state,addr_ch_state')
       /\ awvalid' = 1
       /\ UNCHANGED wvalid
    \/ /\ wvalid = 0
       /\ StartDataChOnly(tran_seq,tran_seq',wdata',wid',data_ch_state,data_ch_state')
       /\ wvalid' = 1
       /\ UNCHANGED awvalid
    \/ /\ awvalid = 0
       /\ wvalid = 0
       /\ StartBothCh(tran_seq,tran_seq',wdata',awaddr',awid',wid',addr_ch_state,addr_ch_state',data_ch_state,data_ch_state') 
       /\ wvalid' = 1
       /\ awvalid' = 1
       


Done ==
    /\ ChDone(tran_seq
                    ,tran_seq'
                    ,addr_ch_state
                    ,addr_ch_state'
                    ,data_ch_state
                    ,data_ch_state'
                    ,wvalid = 1 /\ wready = 1
                    ,awvalid = 1 /\ awready = 1
             )
    /\ \/ <<awvalid,awready>> = <<1,1>>
       \/ <<wvalid,wready>> = <<1,1>>        


AxiBasicInit == 
     /\ TypeOK
     /\ awvalid = 0
     /\ awready = 0
     /\ wvalid = 0
     /\ wready = 0
     /\ awaddr \in ADDR_DONT_CARE
     /\ wdata \in DATA_DONT_CARE
     /\ wid \in ID_DONT_CARE
     /\ awid \in ID_DONT_CARE
     /\ Len(tran_seq) >= 1
     /\ Len(tran_seq) <= MAX_TRAN_LEN
     /\ \E x \in 1..MAX_TRAN_LEN : tran_seq \in [1..x -> tranType] 
     /\ \A y \in DOMAIN tran_seq : tran_seq[y].tran_id = y
     /\ addr_ch_state = [state |-> "Idle",tran_idx |-> DONT_CARE_TRAN_IDX]
     /\ data_ch_state = [state |-> "Idle",tran_idx |-> DONT_CARE_TRAN_IDX]


AxiBasicNext ==
    /\ \/ Start
       \/ Done
       \/ /\ awvalid = 0
          /\ wvalid = 0
          /\ UNCHANGED tran_seq
          /\ <<wready,awready>> \in [1..2 -> {0,1}]
    /\ ADDR_WR_CH!Next
    /\ DATA_WR_CH!Next  


AxiBasicSpec == AxiBasicInit /\ [][AxiBasicNext]_vars /\ WF_vars(AxiBasicNext)
THEOREM AxiBasicSpec => TypeOK
     





 \*IfBecomesBusy == 
 \*   /\ if_state.st = "Idle"
 \*   /\ haveWaitingTran
 \*   /\ if_state' = [if_state EXCEPT ![st] = "Busy",![tran_id] = MinIdle(tran_seq),![addr_ch_st] = "Idle",![data_ch_st] = "Idle" ]
 \*   /\ tran_seq' = [tran_seq EXCEPT ![if_state'[tran_id]].state = "Busy"]
 \*   /\ UNCHANGED <<addr_ch_vars,data_ch_vars>>


\* Becomes idle when all channels have finished their transaction    
\* IfBecomesIdle == 
\*    /\ if_state.st = "Busy"
\*    /\ !(if_state.addr_ch_st = "Done" /\ if_state.data_ch_st = "Done")    

(*
   What is to send transaction
*)


=============================================================================
\* Modification History
\* Last modified Sun May 17 17:28:19 IDT 2020 by levbroido
\* Created Sun May 03 23:52:10 IDT 2020 by levbroido
